use std::{
    any::{Any, TypeId},
    collections::{
        HashMap,
        hash_map::{Iter, IterMut},
    },
    fmt::Display,
};

use anyhow::Result;
use serde::{Deserialize, Serialize};

use crate::{World, get_disjoint_mut::GetDisjointMut};

pub trait EntityComponent: Any + Send + Sync {
    fn enable(&mut self, _world: &mut World) -> Result<()> {
        Ok(())
    }

    fn disable(&mut self, _world: &mut World) -> Result<()> {
        Ok(())
    }

    fn name(&self) -> &str;
}

impl dyn EntityComponent {
    pub fn downcast_owned<T: EntityComponent + 'static>(mut self: Box<Self>) -> Option<T> {
        self.downcast_mut()
            .map(|raw| unsafe { *Box::from_raw(raw as *mut T) })
    }

    pub fn downcast_ref<T: EntityComponent + 'static>(&self) -> Option<&T> {
        (self as &dyn Any).downcast_ref()
    }

    pub fn downcast_mut<T: EntityComponent + 'static>(&mut self) -> Option<&mut T> {
        (self as &mut dyn Any).downcast_mut()
    }
}

pub trait ComponentDebug {
    fn debug_name() -> &'static str;
}

pub trait ComponentCreateInfo<'de>: Serialize + Deserialize<'de> {
    fn to_string(&self) -> Result<String> {
        Ok(ron::to_string(self)?)
    }

    fn from_str(s: &'de str) -> Option<Self> {
        match ron::from_str(s) {
            Ok(t) => Some(t),
            Err(err) => {
                println!("{:?}", err);

                None
            }
        }
    }
}

pub struct TypeMap {
    map: HashMap<TypeId, Box<dyn EntityComponent>>,
}

impl Default for TypeMap {
    fn default() -> Self {
        Self {
            map: HashMap::new(),
        }
    }
}

impl TypeMap {
    pub fn insert<T: EntityComponent + ComponentDebug>(&mut self, value: T) -> Option<T> {
        self.map
            .insert(TypeId::of::<T>(), Box::new(value))
            .map(|any| any.downcast_owned())
            .flatten()
    }

    pub fn insert_type(
        &mut self,
        type_id: TypeId,
        component: Box<dyn EntityComponent>,
    ) -> Option<Box<dyn EntityComponent>> {
        self.map.insert(type_id, component)
    }

    pub fn remove<T: EntityComponent>(&mut self) -> Option<T> {
        self.remove_by_type_id(&TypeId::of::<T>())
            .map(|any| any.downcast_owned())
            .flatten()
    }

    pub fn remove_by_type_id(&mut self, type_id: &TypeId) -> Option<Box<dyn EntityComponent>> {
        self.map.remove(type_id)
    }

    pub fn get<T: EntityComponent + ComponentDebug>(
        &self,
    ) -> std::result::Result<&T, ComponentNotFoundError> {
        self.map
            .get(&TypeId::of::<T>())
            .map(|any| any.downcast_ref().unwrap())
            .ok_or_else(ComponentNotFoundError::component::<T>)
    }

    pub fn get_by_type_id(
        &self,
        type_id: &TypeId,
    ) -> std::result::Result<&Box<dyn EntityComponent>, ComponentNotFoundError> {
        self.map
            .get(type_id)
            .ok_or_else(|| ComponentNotFoundError::type_id(*type_id))
    }

    pub fn get_mut<T: EntityComponent + ComponentDebug>(
        &mut self,
    ) -> std::result::Result<&mut T, ComponentNotFoundError> {
        self.map
            .get_mut(&TypeId::of::<T>())
            .map(|any| any.downcast_mut().unwrap())
            .ok_or_else(ComponentNotFoundError::component::<T>)
    }

    pub fn get_mut_by_type_id(
        &mut self,
        type_id: &TypeId,
    ) -> std::result::Result<&mut Box<dyn EntityComponent>, ComponentNotFoundError> {
        self.map
            .get_mut(type_id)
            .ok_or_else(|| ComponentNotFoundError::type_id(*type_id))
    }

    pub fn contains<T: EntityComponent>(&self) -> bool {
        self.contains_type_id(&TypeId::of::<T>())
    }

    pub fn contains_type_id(&self, type_id: &TypeId) -> bool {
        self.map.contains_key(type_id)
    }

    pub fn iter(&self) -> Iter<'_, TypeId, Box<dyn EntityComponent>> {
        self.map.iter()
    }

    pub fn iter_mut(&mut self) -> IterMut<'_, TypeId, Box<dyn EntityComponent>> {
        self.map.iter_mut()
    }
}

impl<'a, T> GetDisjointMut<'a, &'a mut T> for TypeMap
where
    T: EntityComponent + ComponentDebug,
{
    type Error = ComponentNotFoundError;

    fn get_mut(&'a mut self) -> std::result::Result<&'a mut T, Self::Error> {
        self.map
            .get_mut(&TypeId::of::<T>())
            .map(|any| any.downcast_mut())
            .flatten()
            .ok_or_else(ComponentNotFoundError::component::<T>)
    }
}

macro_rules! impl_get_disjoint_mut {
    ($struct:ident<$($t:ident$(,)?)+>{$error:ident}) => {
        impl<'a, $($t,)+> GetDisjointMut<'a, ($(&'a mut $t,)+)> for $struct
        where
            $(
                $t: EntityComponent + ComponentDebug + 'static,
            )+
        {
            type Error = $error;

            fn get_mut(&'a mut self) -> std::result::Result<($(&'a mut $t,)+), Self::Error> {
                let mut types: std::collections::VecDeque<_>
                    = self.map.get_disjoint_mut([$(&TypeId::of::<$t>(),)+]).into_iter().collect();

                Ok(($(
                    types
                        .pop_front()
                        .flatten()
                        .map(|any| any.downcast_mut())
                        .flatten()
                        .ok_or_else(ComponentNotFoundError::component::<$t>)?,
                )+))
            }
        }

        impl<'a, $($t,)+> GetDisjointMut<'a, ($(Option<&'a mut $t>,)+)> for $struct
        where
            $(
                $t: EntityComponent + ComponentDebug + 'static,
            )+
        {
            type Error = $error;

            fn get_mut(&'a mut self) -> std::result::Result<($(Option<&'a mut $t>,)+), Self::Error> {
                let mut types: std::collections::VecDeque<_>
                    = self.map.get_disjoint_mut([$(&TypeId::of::<$t>(),)+]).into_iter().collect();

                Ok(($(
                    {
                        // dummy needed for macro iteration
                        let _ = stringify!($t);

                        types
                            .pop_front()
                            .flatten()
                            .map(|any| any.downcast_mut())
                            .flatten()
                    },
                )+))
            }
        }
    };
}

impl_get_disjoint_mut!(TypeMap < T, U > { ComponentNotFoundError });
impl_get_disjoint_mut!(TypeMap < T, U, V > { ComponentNotFoundError });
impl_get_disjoint_mut!(TypeMap < T, U, V, W > { ComponentNotFoundError });
impl_get_disjoint_mut!(TypeMap < T, U, V, W, X > { ComponentNotFoundError });
impl_get_disjoint_mut!(TypeMap < T, U, V, W, X, Y > { ComponentNotFoundError });
impl_get_disjoint_mut!(TypeMap < T, U, V, W, X, Y, Z > { ComponentNotFoundError });
#[rustfmt::skip]
impl_get_disjoint_mut!(TypeMap < T, U, V, W, X, Y, Z, A > { ComponentNotFoundError });
#[rustfmt::skip]
impl_get_disjoint_mut!(TypeMap < T, U, V, W, X, Y, Z, A, B > { ComponentNotFoundError });
#[rustfmt::skip]
impl_get_disjoint_mut!(TypeMap < T, U, V, W, X, Y, Z, A, B, C > { ComponentNotFoundError });

#[derive(Debug)]
pub enum ComponentRequestType {
    TypeId(TypeId),
    ComponentName(&'static str),
}

#[derive(Debug)]
pub struct ComponentNotFoundError {
    pub request_type: ComponentRequestType,
}

impl ComponentNotFoundError {
    fn type_id(type_id: TypeId) -> Self {
        Self {
            request_type: ComponentRequestType::TypeId(type_id),
        }
    }

    fn component<T: ComponentDebug>() -> Self {
        Self {
            request_type: ComponentRequestType::ComponentName(T::debug_name()),
        }
    }
}

impl Display for ComponentNotFoundError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "Entity Component ({:?}) not found!", self.request_type)
    }
}

impl std::error::Error for ComponentNotFoundError {}

#[derive(Debug, Default, Clone, Serialize, Deserialize)]
struct Test {
    x: u32,
    y: u32,
    z: u32,
}

impl EntityComponent for Test {
    fn name(&self) -> &str {
        "Test"
    }
}

impl ComponentDebug for Test {
    fn debug_name() -> &'static str {
        "Test"
    }
}

impl PartialEq for Test {
    fn eq(&self, other: &Self) -> bool {
        self.x == other.x && self.y == other.y && self.z == other.z
    }
}

#[test]
fn verify_disjoint_get() {
    let mut map = TypeMap::default();

    let reference = Test { x: 5, y: 20, z: 30 };

    map.insert(reference.clone());
    let old: &mut Test = map.get_mut().unwrap();

    assert_eq!(*old, reference);
}

#[test]
fn verify_insert() {
    let mut map = TypeMap::default();

    let reference = Test { x: 5, y: 20, z: 30 };

    map.insert(reference.clone());
    let old = map.insert(reference.clone()).unwrap();

    assert_eq!(old, reference);
}

#[test]
fn verify_get() {
    let mut map = TypeMap::default();

    let reference = Test { x: 5, y: 20, z: 30 };

    map.insert(reference.clone());

    assert_eq!(map.get::<Test>().unwrap().clone(), reference);
}

#[test]
fn verify_get_mut() {
    let mut map = TypeMap::default();

    let reference = Test { x: 5, y: 20, z: 30 };

    map.insert(reference.clone());

    assert_eq!(map.get_mut::<Test>().unwrap().clone(), reference);
}
